#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
#       Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
#       Earth Sciences Department - Computational Earth Sciences
#     
#     Authors: 
#       Joan Sala Calero (joan.sala@bsc.es)
#       Jesus Pena-Izquierdo (jesus.pena@bsc.es)
#       Ivan Jose Cernicharo Ortiz (ivan.cernicharo@bsc.es)

import sys
import logging
import os
import subprocess

###############################################################################
###############################################################################
# Standard Deviation Developed by Suso
# It Still Needs modifications to work properly
# And it also needs to be implemented in the object workflow properly

import numpy as np
import xarray
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd
from scipy.stats import kurtosis, skew


filespath = '/esarchive/recon/ecmwf/erainterim/6hourly/'
vars = ['tdps','tas','ta100','ta200','ta225','ta250','ta300','ta350','ta400','ta500','ta550','ta600','ta925',
        'sic','snld','swvl1','clt',
        'g200','g250','g50','g500','g850',
        'hurs1','hus850','hus55ml','hus56ml','hus57ml','hus58ml','hus59ml','hus60ml',
        'sfcWind',
        'uas','ua200','ua225','ua250','ua300','ua350','ua400','ua450','ua50','ua500','ua550','ua600','ua700','ua850','ua925',
        'vas','va200','va225','va250','va300','va350','va400','va450','va50','va500','va550','va600','va700','va850','va925']

def Stats(data):
    Mean = np.mean(data)
    Median = np.median(data)
    Std = np.std(data)
    
    #Mad
    diff = np.abs(data - Median)
    Mad = np.median(diff)
    diff_menos = np.abs(data[data <= Median] - Median)
    Mad_menos = np.median(diff_menos)
    diff_mas = np.abs(data[data >= Median] - Median)
    Mad_mas = np.median(diff_mas)
    
    Std_menos =  np.sqrt( np.mean( np.square(data[data<=np.mean(data)] - np.mean(data)))) #np.mean( np.abs(data[data<=np.mean(data)] - np.mean(data)))
    Std_mas = np.sqrt( np.mean( np.square(data[data>=np.mean(data)] - np.mean(data)))) #np.mean( np.abs(data[data>=np.mean(data)] - np.mean(data)))
    Kurt = kurtosis(data)
    Skew = skew(data)
    return np.array([Mean, Median, Kurt, Skew, Std, Std_menos, Std_mas, Mad, Mad_menos, Mad_mas])
    
def minmax_monitor(data, minmax_record):
    if len(minmax_record)==2:
        min_record, max_record = minmax_record
        min_record = min([np.min(data), min_record])
        max_record = max([np.max(data), max_record])
    else:
        min_record = np.min(data)
        max_record = np.max(data)    
    minmax_record = np.array([min_record, max_record])
    return minmax_record

def weight_avg(N__Ni, Var_n__Var_i):
    N, Ni = N__Ni
    Var_n, Var_i = Var_n__Var_i
    
    Var_M = []
    for var_n_k, var_i_k in zip(Var_n, Var_i):
        var_m_k = (var_n_k*N + var_i_k*Ni) / (N + Ni)
        Var_M.append(var_m_k)
    return Var_M

def online_distribution(data, limits_record, distribution_record, bins = 10, OutIn = True):
    n_i, xbin = np.histogram(data, bins = bins, range = limits_record)
    
    if OutIn:
        n_outmin = len(np.where(data < limits_record[0])[0])
        n_outmax = len(np.where(data > limits_record[1])[0])
        n_i[0] = n_i[0] + n_outmax
        n_i[-1] = n_i[-1] + n_outmin
    distribution_record = distribution_record + n_i
    return distribution_record, xbin

def avg_online_stats(filespath, var_ii, bins = 10, OutIn = True):
    
    iMean = 0
    iMed = 1
    iStd = 4
    iMad = 7
    files = sorted(os.listdir(filespath))
    files = [ifile for ifile in files if ifile.startswith(var_ii) and ifile.endswith('.nc')]
    Nt = 0
    minmax_record = []
    
    for ifile in files[:2]:
            
        #Loading single files
        data = xarray.open_dataset(filespath + ifile)[var_ii].values
        Nt_i,Ny,Nx = data.shape
        data = data[~np.isnan(data)].flatten()
        #MinMax record
        minmax_record =  minmax_monitor(data, minmax_record)
        #Stats
        stats_data_i = Stats(data)
        if Nt == 0:
            stats_data_n = [0,] * len(stats_data_i)
            limits_record = minmax_record
            distribution_record = np.zeros((bins,))
        stats_data_n = weight_avg( (Nt, Nt_i), (stats_data_n, stats_data_i))
        distribution_record, xbin = online_distribution(data, limits_record, distribution_record, 
                                                        bins = bins, OutIn = OutIn)
        distribution_record = distribution_record / (Nx*Ny)
        #Proportion between Mean, sd and min/max
        k_sd = np.empty((3,2))
        
        #Proportion between Mads and (Min/Max - Median)
        k_mad = np.empty((3,2))
        for i in range(3):
            if stats_data_n[iMad + i] == 0:
                k_mad[i,:] = np.zeros((2,))
            else:
                k_mad[i,:] = np.abs(minmax_record - stats_data_n[iMed]) / stats_data_n[iMad + i]
            k_sd[i,:] = np.abs(minmax_record - stats_data_n[iMean]) / stats_data_n[iStd + i]
        
        k_sd[1,1] = k_sd[2,1] 
        k_sd = k_sd[:2,:].flatten()
        
        k_mad[1,1] = k_mad[2,1] #First row (Min/Max - Median), second row (Min/Max - Median_menos/Median_mas)
        k_mad = k_mad[:2,:].flatten()        
        Nt += Nt_i
        stats_var = np.concatenate((minmax_record, stats_data_n, k_sd, k_mad))
        print(ifile)
        print( ['Minimum', 'Maximum',
                'Mean', 'Median',
                'Skewness', 'Kurtosis',
                'Std', 'Std_menos', 'Std_mas', 
                'Mad', 'Mad_menos', 'Mad_mas',
                'k_sd_min', 'k_sd_max',
                'k_sd_menos_min', 'k_sd_mas_max',
                'k_mad_min', 'k_mad_max',
                'k_mad_menos_min', 'k_mad_mas_max'])
        print(stats_var)
    return stats_var, distribution_record, xbin

def avg_online_stats_multi(filespath, var_ii, bins = 10, OutIn = True, iMed = 1, iMad = 7, Ni_files = 12):
    iMean = 0
    iMed = 1
    iStd = 4
    iMad = 7
    
    files = sorted(os.listdir(filespath))
    files = [ifile for ifile in files if ifile.startswith(var_ii) and ifile.endswith('.nc')]
    
    Nt = 0
    minmax_record = []
    for i in range(0,len(files), Ni_files):
        
        #Loading multiple files
        os.chdir(filespath)
        ifiles = files[i : i + Ni_files]
        data = xarray.open_mfdataset(ifiles)[var_ii].values
        Nt_i,Ny,Nx = data.shape
        data = data[~np.isnan(data)].flatten()
        
        #MinMax record
        minmax_record =  minmax_monitor(data, minmax_record)
        #Stats
        stats_data_i = Stats(data)
        if Nt == 0:
            stats_data_n = [0,] * len(stats_data_i)
            limits_record = minmax_record
            distribution_record = np.zeros((bins,))
        stats_data_n = weight_avg( (Nt, Nt_i), (stats_data_n, stats_data_i))
        print(limits_record)
        distribution_record, xbin = online_distribution(data, limits_record, distribution_record, 
                                                        bins = bins, OutIn = OutIn)
        distribution_record = distribution_record / (Nx*Ny)
        #Proportion between Mean, sd and min/max
        k_sd = np.empty((3,2))
        
        #Proportion between Mads and (Min/Max - Median)
        k_mad = np.empty((3,2))
        for i in range(3):
            if stats_data_n[iMad + i] == 0:
                k_mad[i,:] = np.zeros((2,))
            else:
                k_mad[i,:] = np.abs(minmax_record - stats_data_n[iMed]) / stats_data_n[iMad + i]
            k_sd[i,:] = np.abs(minmax_record - stats_data_n[iMean]) / stats_data_n[iStd + i]
        
        k_sd[1,1] = k_sd[2,1] 
        k_sd = k_sd[:2,:].flatten()
        
        k_mad[1,1] = k_mad[2,1] #First row (Min/Max - Median), second row (Min/Max - Median_menos/Median_mas)
        k_mad = k_mad[:2,:].flatten()        
        Nt += Nt_i
        stats_var = np.concatenate((minmax_record, stats_data_n, k_sd, k_mad))
        
        print(ifile)
        print( ['Minimum', 'Maximum',
                'Mean', 'Median',
                'Skewness', 'Kurtosis',
                'Std', 'Std_menos', 'Std_mas', 
                'Mad', 'Mad_menos', 'Mad_mas',
                'k_sd_min', 'k_sd_max',
                'k_sd_menos_min', 'k_sd_mas_max',
                'k_mad_min', 'k_mad_max',
                'k_mad_menos_min', 'k_mad_mas_max'])
        print(stats_var)
    return stats_var, distribution_record, xbin

def monitor_multivar(filespath, vars, 
                    outpath = '/home/Earth/jpena/Suso/Coding/Python/',
                    stats_fileout = 'stats.csv',
                    distrib_fileout = 'distributions.csv',
                    bins = 10, OutIn = True, Ni_files = 12):
    cols_name = ['Minimum', 'Maximum',
                'Mean', 'Median',
                'Skewness', 'Kurtosis',
                'Std', 'Std_menos', 'Std_mas', 
                'Mad', 'Mad_menos', 'Mad_mas',
                'k_sd_min', 'k_sd_max',
                'k_sd_menos_min', 'k_sd_mas_max',
                'k_mad_min', 'k_mad_max',
                'k_mad_menos_min', 'k_mad_mas_max']
    stats_vars = pd.DataFrame(columns = cols_name)
    distribution_vars = pd.DataFrame() # = pd.Series()
    for var_i in vars:
        
        i_ = vars[0].find('_')
        if i_>0:
            var_ii = var_i[:i_]
        else:
            var_ii = var_i
        filespath_var_i = filespath + var_i + '/'
        
        print('Reading {}{}'.format(filespath_var_i,var_ii))
        #stats_var_i, distribution_record, xbin = avg_online_stats_multi(filespath_var_i, var_ii, bins = bins, OutIn = OutIn, Ni_files = Ni_files)
        stats_var_i, distribution_record, xbin = avg_online_stats(filespath_var_i, var_ii, bins = bins, OutIn = OutIn)
        
        s_stats_var_i = pd.Series( stats_var_i, name = var_ii, index = cols_name)
        stats_vars = stats_vars.append(s_stats_var_i)
        s_distrib_var_i = pd.Series( distribution_record, name = var_ii) # = pd.Series( distribution_record, index = [var_i]*bins, xbin)
        distribution_vars = distribution_vars.append(s_distrib_var_i) # = distribution_vars.append(s_distrib_var_i)
        print(' ')
        print(var_i + ' added.')
        print(' ')
        
    stats_vars.to_csv(outpath + stats_fileout)
    distribution_vars.to_csv(outpath + distrib_fileout)
    
    return stats_vars, distribution_vars



###############################################################################
###############################################################################
###############################################################################
###############################################################################

from utils.utils_files import *

"""
Class: Valid ranges checks
---
- Checks data to detect outliers
"""
class valid_ranges_test:

    def __init__(self, logdir, path, pattern):
        self.testid = 'test01-standard-compliance'
        self.logdir = logdir 
        self.path = path
        raise Exception("valid_ranges checks are not implemented yet. They will be once their scientific assessment has been agreed ")
        sys.exit()        
        self.files = dir_or_file(path, pattern)
    
    def main(self):
        
        # Initiate logging
        initiate_logging(self.logdir, self.path, self.testid)

        print('WILL BE READY SOON')        

if __name__ == "__main__":
    sys.exit()
